"""
Support for the Automatic platform.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/device_tracker.automatic/
"""
import asyncio
from datetime import timedelta
import logging

import voluptuous as vol

from homeassistant.components.device_tracker import (
    PLATFORM_SCHEMA, ATTR_ATTRIBUTES)
from homeassistant.const import CONF_USERNAME, CONF_PASSWORD
from homeassistant.helpers.aiohttp_client import async_get_clientsession
import homeassistant.helpers.config_validation as cv
from homeassistant.helpers.event import async_track_time_interval

REQUIREMENTS = ['aioautomatic==0.2.1']

_LOGGER = logging.getLogger(__name__)

CONF_CLIENT_ID = 'client_id'
CONF_SECRET = 'secret'
CONF_DEVICES = 'devices'

DEFAULT_TIMEOUT = 5

SCOPE = ['location', 'vehicle:profile', 'trip']

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Required(CONF_CLIENT_ID): cv.string,
    vol.Required(CONF_SECRET): cv.string,
    vol.Required(CONF_USERNAME): cv.string,
    vol.Required(CONF_PASSWORD): cv.string,
    vol.Optional(CONF_DEVICES, default=None): vol.All(
        cv.ensure_list, [cv.string])
})


@asyncio.coroutine
def async_setup_scanner(hass, config, async_see, discovery_info=None):
    """Validate the configuration and return an Automatic scanner."""
    import aioautomatic

    client = aioautomatic.Client(
        client_id=config[CONF_CLIENT_ID],
        client_secret=config[CONF_SECRET],
        client_session=async_get_clientsession(hass),
        request_kwargs={'timeout': DEFAULT_TIMEOUT})
    try:
        session = yield from client.create_session_from_password(
            SCOPE, config[CONF_USERNAME], config[CONF_PASSWORD])
        data = AutomaticData(hass, session, config[CONF_DEVICES], async_see)
    except aioautomatic.exceptions.AutomaticError as err:
        _LOGGER.error(str(err))
        return False

    yield from data.update()
    return True


class AutomaticData(object):
    """A class representing an Automatic cloud service connection."""

    def __init__(self, hass, session, devices, async_see):
        """Initialize the automatic device scanner."""
        self.hass = hass
        self.devices = devices
        self.session = session
        self.async_see = async_see

        async_track_time_interval(hass, self.update, timedelta(seconds=30))

    @asyncio.coroutine
    def update(self, now=None):
        """Update the device info."""
        import aioautomatic

        _LOGGER.debug('Updating devices %s', now)

        try:
            vehicles = yield from self.session.get_vehicles()
        except aioautomatic.exceptions.AutomaticError as err:
            _LOGGER.error(str(err))
            return False

        for vehicle in vehicles:
            name = vehicle.display_name
            if name is None:
                name = ' '.join(filter(None, (
                    str(vehicle.year), vehicle.make, vehicle.model)))

            if self.devices is not None and name not in self.devices:
                continue

            self.hass.async_add_job(self.update_vehicle(vehicle, name))

    @asyncio.coroutine
    def update_vehicle(self, vehicle, name):
        """Updated the specified vehicle's data."""
        import aioautomatic

        kwargs = {
            'dev_id': vehicle.id,
            'host_name': name,
            'mac': vehicle.id,
            ATTR_ATTRIBUTES: {
                'fuel_level': vehicle.fuel_level_percent,
                }
        }

        trips = []
        try:
            # Get the most recent trip for this vehicle
            trips = yield from self.session.get_trips(
                vehicle=vehicle.id, limit=1)
        except aioautomatic.exceptions.AutomaticError as err:
            _LOGGER.error(str(err))

        if trips:
            end_location = trips[0].end_location
            kwargs['gps'] = (end_location.lat, end_location.lon)
            kwargs['gps_accuracy'] = end_location.accuracy_m

        yield from self.async_see(**kwargs)



"""Support for light effects for the LIFX light platform."""
import logging
import asyncio
import random
from os import path

import voluptuous as vol

from homeassistant.components.light import (
    DOMAIN, ATTR_BRIGHTNESS, ATTR_COLOR_NAME, ATTR_RGB_COLOR, ATTR_EFFECT)
from homeassistant.config import load_yaml_config_file
from homeassistant.const import (ATTR_ENTITY_ID)
from homeassistant.helpers.service import extract_entity_ids
import homeassistant.helpers.config_validation as cv

_LOGGER = logging.getLogger(__name__)

SERVICE_EFFECT_BREATHE = 'lifx_effect_breathe'
SERVICE_EFFECT_PULSE = 'lifx_effect_pulse'
SERVICE_EFFECT_COLORLOOP = 'lifx_effect_colorloop'
SERVICE_EFFECT_STOP = 'lifx_effect_stop'

ATTR_POWER_ON = 'power_on'
ATTR_PERIOD = 'period'
ATTR_CYCLES = 'cycles'
ATTR_SPREAD = 'spread'
ATTR_CHANGE = 'change'

# aiolifx waveform modes
WAVEFORM_SINE = 1
WAVEFORM_PULSE = 4

LIFX_EFFECT_SCHEMA = vol.Schema({
    vol.Optional(ATTR_ENTITY_ID): cv.entity_ids,
    vol.Optional(ATTR_POWER_ON, default=True): cv.boolean,
})

LIFX_EFFECT_BREATHE_SCHEMA = LIFX_EFFECT_SCHEMA.extend({
    ATTR_BRIGHTNESS: vol.All(vol.Coerce(int), vol.Clamp(min=0, max=255)),
    ATTR_COLOR_NAME: cv.string,
    ATTR_RGB_COLOR: vol.All(vol.ExactSequence((cv.byte, cv.byte, cv.byte)),
                            vol.Coerce(tuple)),
    vol.Optional(ATTR_PERIOD, default=1.0): vol.All(vol.Coerce(float),
                                                    vol.Range(min=0.05)),
    vol.Optional(ATTR_CYCLES, default=1.0): vol.All(vol.Coerce(float),
                                                    vol.Range(min=1)),
})

LIFX_EFFECT_PULSE_SCHEMA = LIFX_EFFECT_BREATHE_SCHEMA

LIFX_EFFECT_COLORLOOP_SCHEMA = LIFX_EFFECT_SCHEMA.extend({
    ATTR_BRIGHTNESS: vol.All(vol.Coerce(int), vol.Clamp(min=0, max=255)),
    vol.Optional(ATTR_PERIOD, default=60): vol.All(vol.Coerce(float),
                                                   vol.Clamp(min=1)),
    vol.Optional(ATTR_CHANGE, default=20): vol.All(vol.Coerce(float),
                                                   vol.Clamp(min=0, max=360)),
    vol.Optional(ATTR_SPREAD, default=30): vol.All(vol.Coerce(float),
                                                   vol.Clamp(min=0, max=360)),
})

LIFX_EFFECT_STOP_SCHEMA = vol.Schema({
    vol.Optional(ATTR_ENTITY_ID): cv.entity_ids,
    vol.Optional(ATTR_POWER_ON, default=False): cv.boolean,
})


def setup(hass, lifx_manager):
    """Register the LIFX effects as hass service calls."""
    @asyncio.coroutine
    def async_service_handle(service):
        """Internal func for applying a service."""
        entity_ids = extract_entity_ids(hass, service)
        if entity_ids:
            devices = [entity for entity in lifx_manager.entities.values()
                       if entity.entity_id in entity_ids]
        else:
            devices = list(lifx_manager.entities.values())

        if devices:
            yield from start_effect(hass, devices,
                                    service.service, **service.data)

    descriptions = load_yaml_config_file(
        path.join(path.dirname(__file__), 'services.yaml'))

    hass.services.async_register(
        DOMAIN, SERVICE_EFFECT_BREATHE, async_service_handle,
        descriptions.get(SERVICE_EFFECT_BREATHE),
        schema=LIFX_EFFECT_BREATHE_SCHEMA)

    hass.services.async_register(
        DOMAIN, SERVICE_EFFECT_PULSE, async_service_handle,
        descriptions.get(SERVICE_EFFECT_PULSE),
        schema=LIFX_EFFECT_PULSE_SCHEMA)

    hass.services.async_register(
        DOMAIN, SERVICE_EFFECT_COLORLOOP, async_service_handle,
        descriptions.get(SERVICE_EFFECT_COLORLOOP),
        schema=LIFX_EFFECT_COLORLOOP_SCHEMA)

    hass.services.async_register(
        DOMAIN, SERVICE_EFFECT_STOP, async_service_handle,
        descriptions.get(SERVICE_EFFECT_STOP),
        schema=LIFX_EFFECT_STOP_SCHEMA)


@asyncio.coroutine
def start_effect(hass, devices, service, **data):
    """Start a light effect."""
    tasks = []
    for light in devices:
        tasks.append(hass.async_add_job(light.stop_effect()))
    yield from asyncio.wait(tasks, loop=hass.loop)

    if service in SERVICE_EFFECT_BREATHE:
        effect = LIFXEffectBreathe(hass, devices)
    elif service in SERVICE_EFFECT_PULSE:
        effect = LIFXEffectPulse(hass, devices)
    elif service == SERVICE_EFFECT_COLORLOOP:
        effect = LIFXEffectColorloop(hass, devices)
    elif service == SERVICE_EFFECT_STOP:
        effect = LIFXEffectStop(hass, devices)

    hass.async_add_job(effect.async_perform(**data))


@asyncio.coroutine
def default_effect(light, **kwargs):
    """Start an effect with default parameters."""
    service = kwargs[ATTR_EFFECT]
    data = {
        ATTR_ENTITY_ID: light.entity_id,
    }
    if service in (SERVICE_EFFECT_BREATHE, SERVICE_EFFECT_PULSE):
        data[ATTR_RGB_COLOR] = [
            random.randint(1, 127),
            random.randint(1, 127),
            random.randint(1, 127),
        ]
        data[ATTR_BRIGHTNESS] = 255
    yield from light.hass.services.async_call(DOMAIN, service, data)


def effect_list():
    """Return the list of supported effects."""
    return [
        SERVICE_EFFECT_COLORLOOP,
        SERVICE_EFFECT_BREATHE,
        SERVICE_EFFECT_PULSE,
        SERVICE_EFFECT_STOP,
    ]


class LIFXEffectData(object):
    """Structure describing a running effect."""

    def __init__(self, effect, power, color):
        """Initialize data structure."""
        self.effect = effect
        self.power = power
        self.color = color


class LIFXEffect(object):
    """Representation of a light effect running on a number of lights."""

    def __init__(self, hass, lights):
        """Initialize the effect."""
        self.hass = hass
        self.lights = lights

    @asyncio.coroutine
    def async_perform(self, **kwargs):
        """Do common setup and play the effect."""
        yield from self.async_setup(**kwargs)
        yield from self.async_play(**kwargs)

    @asyncio.coroutine
    def async_setup(self, **kwargs):
        """Prepare all lights for the effect."""
        for light in self.lights:
            yield from light.refresh_state()
            if not light.device:
                self.lights.remove(light)
            else:
                light.effect_data = LIFXEffectData(
                    self, light.is_on, light.device.color)

                # Temporarily turn on power for the effect to be visible
                if kwargs[ATTR_POWER_ON] and not light.is_on:
                    hsbk = self.from_poweroff_hsbk(light, **kwargs)
                    light.device.set_color(hsbk)
                    light.device.set_power(True)

    # pylint: disable=no-self-use
    @asyncio.coroutine
    def async_play(self, **kwargs):
        """Play the effect."""
        yield None

    @asyncio.coroutine
    def async_restore(self, light):
        """Restore to the original state (if we are still running)."""
        if light.effect_data:
            if light.effect_data.effect == self:
                if light.device and not light.effect_data.power:
                    light.device.set_power(False)
                    yield from asyncio.sleep(0.5)
                if light.device:
                    light.device.set_color(light.effect_data.color)
                    yield from asyncio.sleep(0.5)
                light.effect_data = None
            self.lights.remove(light)

    def from_poweroff_hsbk(self, light, **kwargs):
        """The initial color when starting from a powered off state."""
        return None


class LIFXEffectBreathe(LIFXEffect):
    """Representation of a breathe effect."""

    def __init__(self, hass, lights):
        """Initialize the breathe effect."""
        super(LIFXEffectBreathe, self).__init__(hass, lights)
        self.name = SERVICE_EFFECT_BREATHE
        self.waveform = WAVEFORM_SINE

    @asyncio.coroutine
    def async_play(self, **kwargs):
        """Play the effect on all lights."""
        for light in self.lights:
            self.hass.async_add_job(self.async_light_play(light, **kwargs))

    @asyncio.coroutine
    def async_light_play(self, light, **kwargs):
        """Play a light effect on the bulb."""
        period = kwargs[ATTR_PERIOD]
        cycles = kwargs[ATTR_CYCLES]
        hsbk, _ = light.find_hsbk(**kwargs)

        # Start the effect
        args = {
            'transient': 1,
            'color': hsbk,
            'period': int(period*1000),
            'cycles': cycles,
            'duty_cycle': 0,
            'waveform': self.waveform,
        }
        light.device.set_waveform(args)

        # Wait for completion and restore the initial state
        yield from asyncio.sleep(period*cycles)
        yield from self.async_restore(light)

    def from_poweroff_hsbk(self, light, **kwargs):
        """Initial color is the target color, but no brightness."""
        hsbk, _ = light.find_hsbk(**kwargs)
        return [hsbk[0], hsbk[1], 0, hsbk[2]]


class LIFXEffectPulse(LIFXEffectBreathe):
    """Representation of a pulse effect."""

    def __init__(self, hass, lights):
        """Initialize the pulse effect."""
        super(LIFXEffectPulse, self).__init__(hass, lights)
        self.name = SERVICE_EFFECT_PULSE
        self.waveform = WAVEFORM_PULSE


class LIFXEffectColorloop(LIFXEffect):
    """Representation of a colorloop effect."""

    def __init__(self, hass, lights):
        """Initialize the colorloop effect."""
        super(LIFXEffectColorloop, self).__init__(hass, lights)
        self.name = SERVICE_EFFECT_COLORLOOP

    @asyncio.coroutine
    def async_play(self, **kwargs):
        """Play the effect on all lights."""
        period = kwargs[ATTR_PERIOD]
        spread = kwargs[ATTR_SPREAD]
        change = kwargs[ATTR_CHANGE]
        direction = 1 if random.randint(0, 1) else -1

        # Random start
        hue = random.randint(0, 359)

        while self.lights:
            hue = (hue + direction*change) % 360

            random.shuffle(self.lights)
            lhue = hue

            transition = int(1000 * random.uniform(period/2, period))
            for light in self.lights:
                if spread > 0:
                    transition = int(1000 * random.uniform(period/2, period))

                if ATTR_BRIGHTNESS in kwargs:
                    brightness = int(65535/255*kwargs[ATTR_BRIGHTNESS])
                else:
                    brightness = light.effect_data.color[2]

                hsbk = [
                    int(65535/359*lhue),
                    int(random.uniform(0.8, 1.0)*65535),
                    brightness,
                    4000,
                ]
                light.device.set_color(hsbk, None, transition)

                # Adjust the next light so the full spread is used
                if len(self.lights) > 1:
                    lhue = (lhue + spread/(len(self.lights)-1)) % 360

            yield from asyncio.sleep(period)

    def from_poweroff_hsbk(self, light, **kwargs):
        """Start from a random hue."""
        return [random.randint(0, 65535), 65535, 0, 4000]


class LIFXEffectStop(LIFXEffect):
    """A no-op effect, but starting it will stop an existing effect."""

    def __init__(self, hass, lights):
        """Initialize the stop effect."""
        super(LIFXEffectStop, self).__init__(hass, lights)
        self.name = SERVICE_EFFECT_STOP

    @asyncio.coroutine
    def async_perform(self, **kwargs):
        """Do nothing."""
        yield None



"""
Support for AlarmDecoder devices.

For more details about this component, please refer to the documentation at
https://home-assistant.io/components/alarmdecoder/
"""
import asyncio
import logging

import voluptuous as vol
import homeassistant.helpers.config_validation as cv

from homeassistant.core import callback
from homeassistant.const import EVENT_HOMEASSISTANT_STOP

from homeassistant.helpers.discovery import async_load_platform
from homeassistant.helpers.dispatcher import async_dispatcher_send

REQUIREMENTS = ['alarmdecoder==0.12.1.0']

_LOGGER = logging.getLogger(__name__)

DOMAIN = 'alarmdecoder'

DATA_AD = 'alarmdecoder'


CONF_DEVICE = 'device'
CONF_DEVICE_TYPE = 'type'
CONF_DEVICE_HOST = 'host'
CONF_DEVICE_PORT = 'port'
CONF_DEVICE_PATH = 'path'
CONF_DEVICE_BAUD = 'baudrate'

CONF_ZONES = 'zones'
CONF_ZONE_NAME = 'name'
CONF_ZONE_TYPE = 'type'

CONF_PANEL_DISPLAY = 'panel_display'

DEFAULT_DEVICE_TYPE = 'socket'
DEFAULT_DEVICE_HOST = 'localhost'
DEFAULT_DEVICE_PORT = 10000
DEFAULT_DEVICE_PATH = '/dev/ttyUSB0'
DEFAULT_DEVICE_BAUD = 115200

DEFAULT_PANEL_DISPLAY = False

DEFAULT_ZONE_TYPE = 'opening'

SIGNAL_PANEL_MESSAGE = 'alarmdecoder.panel_message'
SIGNAL_PANEL_ARM_AWAY = 'alarmdecoder.panel_arm_away'
SIGNAL_PANEL_ARM_HOME = 'alarmdecoder.panel_arm_home'
SIGNAL_PANEL_DISARM = 'alarmdecoder.panel_disarm'

SIGNAL_ZONE_FAULT = 'alarmdecoder.zone_fault'
SIGNAL_ZONE_RESTORE = 'alarmdecoder.zone_restore'

DEVICE_SOCKET_SCHEMA = vol.Schema({
    vol.Required(CONF_DEVICE_TYPE): 'socket',
    vol.Optional(CONF_DEVICE_HOST, default=DEFAULT_DEVICE_HOST): cv.string,
    vol.Optional(CONF_DEVICE_PORT, default=DEFAULT_DEVICE_PORT): cv.port})

DEVICE_SERIAL_SCHEMA = vol.Schema({
    vol.Required(CONF_DEVICE_TYPE): 'serial',
    vol.Optional(CONF_DEVICE_PATH, default=DEFAULT_DEVICE_PATH): cv.string,
    vol.Optional(CONF_DEVICE_BAUD, default=DEFAULT_DEVICE_BAUD): cv.string})

DEVICE_USB_SCHEMA = vol.Schema({
    vol.Required(CONF_DEVICE_TYPE): 'usb'})

ZONE_SCHEMA = vol.Schema({
    vol.Required(CONF_ZONE_NAME): cv.string,
    vol.Optional(CONF_ZONE_TYPE, default=DEFAULT_ZONE_TYPE): cv.string})

CONFIG_SCHEMA = vol.Schema({
    DOMAIN: vol.Schema({
        vol.Required(CONF_DEVICE): vol.Any(DEVICE_SOCKET_SCHEMA,
                                           DEVICE_SERIAL_SCHEMA,
                                           DEVICE_USB_SCHEMA),
        vol.Optional(CONF_PANEL_DISPLAY,
                     default=DEFAULT_PANEL_DISPLAY): cv.boolean,
        vol.Optional(CONF_ZONES): {vol.Coerce(int): ZONE_SCHEMA},
    }),
}, extra=vol.ALLOW_EXTRA)


@asyncio.coroutine
def async_setup(hass, config):
    """Common setup for AlarmDecoder devices."""
    from alarmdecoder import AlarmDecoder
    from alarmdecoder.devices import (SocketDevice, SerialDevice, USBDevice)

    conf = config.get(DOMAIN)

    device = conf.get(CONF_DEVICE)
    display = conf.get(CONF_PANEL_DISPLAY)
    zones = conf.get(CONF_ZONES)

    device_type = device.get(CONF_DEVICE_TYPE)
    host = DEFAULT_DEVICE_HOST
    port = DEFAULT_DEVICE_PORT
    path = DEFAULT_DEVICE_PATH
    baud = DEFAULT_DEVICE_BAUD

    sync_connect = asyncio.Future(loop=hass.loop)

    def handle_open(device):
        """Callback for a successful connection."""
        _LOGGER.info("Established a connection with the alarmdecoder.")
        hass.bus.async_listen_once(EVENT_HOMEASSISTANT_STOP, stop_alarmdecoder)
        sync_connect.set_result(True)

    @callback
    def stop_alarmdecoder(event):
        """Callback to handle shutdown alarmdecoder."""
        _LOGGER.debug("Shutting down alarmdecoder.")
        controller.close()

    @callback
    def handle_message(sender, message):
        """Callback to handle message from alarmdecoder."""
        async_dispatcher_send(hass, SIGNAL_PANEL_MESSAGE, message)

    def zone_fault_callback(sender, zone):
        """Callback to handle zone fault from alarmdecoder."""
        async_dispatcher_send(hass, SIGNAL_ZONE_FAULT, zone)

    def zone_restore_callback(sender, zone):
        """Callback to handle zone restore from alarmdecoder."""
        async_dispatcher_send(hass, SIGNAL_ZONE_RESTORE, zone)

    controller = False
    if device_type == 'socket':
        host = device.get(CONF_DEVICE_HOST)
        port = device.get(CONF_DEVICE_PORT)
        controller = AlarmDecoder(SocketDevice(interface=(host, port)))
    elif device_type == 'serial':
        path = device.get(CONF_DEVICE_PATH)
        baud = device.get(CONF_DEVICE_BAUD)
        controller = AlarmDecoder(SerialDevice(interface=path))
    elif device_type == 'usb':
        AlarmDecoder(USBDevice.find())
        return False

    controller.on_open += handle_open
    controller.on_message += handle_message
    controller.on_zone_fault += zone_fault_callback
    controller.on_zone_restore += zone_restore_callback

    hass.data[DATA_AD] = controller

    controller.open(baud)

    result = yield from sync_connect

    if not result:
        return False

    hass.async_add_job(async_load_platform(hass, 'alarm_control_panel', DOMAIN,
                                           conf, config))

    if zones:
        hass.async_add_job(async_load_platform(
            hass, 'binary_sensor', DOMAIN, {CONF_ZONES: zones}, config))

    if display:
        hass.async_add_job(async_load_platform(hass, 'sensor', DOMAIN,
                                               conf, config))

    return True



"""Interfaces with TotalConnect alarm control panels."""
import logging

import voluptuous as vol

import homeassistant.components.alarm_control_panel as alarm
from homeassistant.components.alarm_control_panel import PLATFORM_SCHEMA
from homeassistant.const import (
    CONF_PASSWORD, CONF_USERNAME, STATE_ALARM_ARMED_AWAY,
    STATE_ALARM_ARMED_HOME, STATE_ALARM_DISARMED, STATE_UNKNOWN,
    CONF_NAME)
import homeassistant.helpers.config_validation as cv

REQUIREMENTS = ['total_connect_client==0.7']

_LOGGER = logging.getLogger(__name__)

DEFAULT_NAME = 'Total Connect'

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Required(CONF_PASSWORD): cv.string,
    vol.Required(CONF_USERNAME): cv.string,
    vol.Optional(CONF_NAME, default=DEFAULT_NAME): cv.string,
})


def setup_platform(hass, config, add_devices, discovery_info=None):
    """Setup a TotalConnect control panel."""
    name = config.get(CONF_NAME)
    username = config.get(CONF_USERNAME)
    password = config.get(CONF_PASSWORD)

    total_connect = TotalConnect(name, username, password)
    add_devices([total_connect], True)


class TotalConnect(alarm.AlarmControlPanel):
    """Represent an TotalConnect status."""

    def __init__(self, name, username, password):
        """Initialize the TotalConnect status."""
        from total_connect_client import TotalConnectClient

        _LOGGER.debug('Setting up TotalConnect...')
        self._name = name
        self._username = username
        self._password = password
        self._state = STATE_UNKNOWN
        self._client = TotalConnectClient.TotalConnectClient(username,
                                                             password)

    @property
    def name(self):
        """Return the name of the device."""
        return self._name

    @property
    def state(self):
        """Return the state of the device."""
        return self._state

    def update(self):
        """Return the state of the device."""
        status = self._client.get_armed_status()

        if status == self._client.DISARMED:
            state = STATE_ALARM_DISARMED
        elif status == self._client.ARMED_STAY:
            state = STATE_ALARM_ARMED_HOME
        elif status == self._client.ARMED_AWAY:
            state = STATE_ALARM_ARMED_AWAY
        else:
            state = STATE_UNKNOWN

        self._state = state

    def alarm_disarm(self, code=None):
        """Send disarm command."""
        self._client.disarm()

    def alarm_arm_home(self, code=None):
        """Send arm home command."""
        self._client.arm_stay()

    def alarm_arm_away(self, code=None):
        """Send arm away command."""
        self._client.arm_away()



"""
Support for the LIFX platform that implements lights.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/light.lifx/
"""
import colorsys
import logging
import asyncio
import sys
from functools import partial
from datetime import timedelta
import async_timeout

import voluptuous as vol

from homeassistant.components.light import (
    Light, PLATFORM_SCHEMA, ATTR_BRIGHTNESS, ATTR_COLOR_NAME, ATTR_RGB_COLOR,
    ATTR_XY_COLOR, ATTR_COLOR_TEMP, ATTR_TRANSITION, ATTR_EFFECT,
    SUPPORT_BRIGHTNESS, SUPPORT_COLOR_TEMP, SUPPORT_RGB_COLOR,
    SUPPORT_XY_COLOR, SUPPORT_TRANSITION, SUPPORT_EFFECT)
from homeassistant.util.color import (
    color_temperature_mired_to_kelvin, color_temperature_kelvin_to_mired)
from homeassistant import util
from homeassistant.core import callback
from homeassistant.helpers.event import async_track_point_in_utc_time
import homeassistant.helpers.config_validation as cv
import homeassistant.util.color as color_util

from . import effects as lifx_effects

_LOGGER = logging.getLogger(__name__)

REQUIREMENTS = ['aiolifx==0.4.4']

UDP_BROADCAST_PORT = 56700

# Delay (in ms) expected for changes to take effect in the physical bulb
BULB_LATENCY = 500

CONF_SERVER = 'server'

ATTR_HSBK = 'hsbk'

BYTE_MAX = 255
SHORT_MAX = 65535

SUPPORT_LIFX = (SUPPORT_BRIGHTNESS | SUPPORT_COLOR_TEMP | SUPPORT_RGB_COLOR |
                SUPPORT_XY_COLOR | SUPPORT_TRANSITION | SUPPORT_EFFECT)

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Optional(CONF_SERVER, default='0.0.0.0'): cv.string,
})


@asyncio.coroutine
def async_setup_platform(hass, config, async_add_devices, discovery_info=None):
    """Setup the LIFX platform."""
    import aiolifx

    if sys.platform == 'win32':
        _LOGGER.warning('The lifx platform is known to not work on Windows. '
                        'Consider using the lifx_legacy platform instead.')

    server_addr = config.get(CONF_SERVER)

    lifx_manager = LIFXManager(hass, async_add_devices)

    coro = hass.loop.create_datagram_endpoint(
        partial(aiolifx.LifxDiscovery, hass.loop, lifx_manager),
        local_addr=(server_addr, UDP_BROADCAST_PORT))

    hass.async_add_job(coro)

    lifx_effects.setup(hass, lifx_manager)

    return True


class LIFXManager(object):
    """Representation of all known LIFX entities."""

    def __init__(self, hass, async_add_devices):
        """Initialize the light."""
        self.entities = {}
        self.hass = hass
        self.async_add_devices = async_add_devices

    @callback
    def register(self, device):
        """Callback for newly detected bulb."""
        if device.mac_addr in self.entities:
            entity = self.entities[device.mac_addr]
            entity.device = device
            _LOGGER.debug("%s register AGAIN", entity.who)
            self.hass.async_add_job(entity.async_update_ha_state())
        else:
            _LOGGER.debug("%s register NEW", device.ip_addr)
            device.get_color(self.ready)

    @callback
    def ready(self, device, msg):
        """Callback that adds the device once all data is retrieved."""
        entity = LIFXLight(device)
        _LOGGER.debug("%s register READY", entity.who)
        self.entities[device.mac_addr] = entity
        self.async_add_devices([entity])

    @callback
    def unregister(self, device):
        """Callback for disappearing bulb."""
        if device.mac_addr in self.entities:
            entity = self.entities[device.mac_addr]
            _LOGGER.debug("%s unregister", entity.who)
            entity.device = None
            self.hass.async_add_job(entity.async_update_ha_state())


class AwaitAioLIFX:
    """Wait for an aiolifx callback and return the message."""

    def __init__(self, light):
        """Initialize the wrapper."""
        self.light = light
        self.device = None
        self.message = None
        self.event = asyncio.Event()

    @callback
    def callback(self, device, message):
        """Callback that aiolifx invokes when the response is received."""
        self.device = device
        self.message = message
        self.event.set()

    @asyncio.coroutine
    def wait(self, method):
        """Call an aiolifx method and wait for its response or a timeout."""
        self.event.clear()
        method(self.callback)

        while self.light.available and not self.event.is_set():
            try:
                with async_timeout.timeout(1.0, loop=self.light.hass.loop):
                    yield from self.event.wait()
            except asyncio.TimeoutError:
                pass

        return self.message


def convert_rgb_to_hsv(rgb):
    """Convert Home Assistant RGB values to HSV values."""
    red, green, blue = [_ / BYTE_MAX for _ in rgb]

    hue, saturation, brightness = colorsys.rgb_to_hsv(red, green, blue)

    return [int(hue * SHORT_MAX),
            int(saturation * SHORT_MAX),
            int(brightness * SHORT_MAX)]


class LIFXLight(Light):
    """Representation of a LIFX light."""

    def __init__(self, device):
        """Initialize the light."""
        self.device = device
        self.blocker = None
        self.effect_data = None
        self.postponed_update = None
        self._name = device.label
        self.set_power(device.power_level)
        self.set_color(*device.color)

    @property
    def available(self):
        """Return the availability of the device."""
        return self.device is not None

    @property
    def name(self):
        """Return the name of the device."""
        return self._name

    @property
    def who(self):
        """Return a string identifying the device."""
        ip_addr = '-'
        if self.device:
            ip_addr = self.device.ip_addr[0]
        return "%s (%s)" % (ip_addr, self.name)

    @property
    def rgb_color(self):
        """Return the RGB value."""
        _LOGGER.debug(
            "rgb_color: [%d %d %d]", self._rgb[0], self._rgb[1], self._rgb[2])
        return self._rgb

    @property
    def brightness(self):
        """Return the brightness of this light between 0..255."""
        brightness = int(self._bri / (BYTE_MAX + 1))
        _LOGGER.debug("brightness: %d", brightness)
        return brightness

    @property
    def color_temp(self):
        """Return the color temperature."""
        temperature = color_temperature_kelvin_to_mired(self._kel)

        _LOGGER.debug("color_temp: %d", temperature)
        return temperature

    @property
    def is_on(self):
        """Return true if device is on."""
        _LOGGER.debug("is_on: %d", self._power)
        return self._power != 0

    @property
    def effect(self):
        """Return the currently running effect."""
        return self.effect_data.effect.name if self.effect_data else None

    @property
    def supported_features(self):
        """Flag supported features."""
        return SUPPORT_LIFX

    @property
    def effect_list(self):
        """Return the list of supported effects."""
        return lifx_effects.effect_list()

    @callback
    def update_after_transition(self, now):
        """Request new status after completion of the last transition."""
        self.postponed_update = None
        self.hass.async_add_job(self.async_update_ha_state(force_refresh=True))

    @callback
    def unblock_updates(self, now):
        """Allow async_update after the new state has settled on the bulb."""
        self.blocker = None
        self.hass.async_add_job(self.async_update_ha_state(force_refresh=True))

    def update_later(self, when):
        """Block immediate update requests and schedule one for later."""
        if self.blocker:
            self.blocker()
        self.blocker = async_track_point_in_utc_time(
            self.hass, self.unblock_updates,
            util.dt.utcnow() + timedelta(milliseconds=BULB_LATENCY))

        if self.postponed_update:
            self.postponed_update()
        if when > BULB_LATENCY:
            self.postponed_update = async_track_point_in_utc_time(
                self.hass, self.update_after_transition,
                util.dt.utcnow() + timedelta(milliseconds=when+BULB_LATENCY))

    @asyncio.coroutine
    def async_turn_on(self, **kwargs):
        """Turn the device on."""
        yield from self.stop_effect()

        if ATTR_EFFECT in kwargs:
            yield from lifx_effects.default_effect(self, **kwargs)
            return

        if ATTR_TRANSITION in kwargs:
            fade = int(kwargs[ATTR_TRANSITION] * 1000)
        else:
            fade = 0

        hsbk, changed_color = self.find_hsbk(**kwargs)
        _LOGGER.debug("turn_on: %s (%d) %d %d %d %d %d",
                      self.who, self._power, fade, *hsbk)

        if self._power == 0:
            if changed_color:
                self.device.set_color(hsbk, None, 0)
            self.device.set_power(True, None, fade)
        else:
            self.device.set_power(True, None, 0)     # racing for power status
            if changed_color:
                self.device.set_color(hsbk, None, fade)

        self.update_later(0)
        if fade < BULB_LATENCY:
            self.set_power(1)
            self.set_color(*hsbk)

    @asyncio.coroutine
    def async_turn_off(self, **kwargs):
        """Turn the device off."""
        yield from self.stop_effect()

        if ATTR_TRANSITION in kwargs:
            fade = int(kwargs[ATTR_TRANSITION] * 1000)
        else:
            fade = 0

        self.device.set_power(False, None, fade)

        self.update_later(fade)
        if fade < BULB_LATENCY:
            self.set_power(0)

    @asyncio.coroutine
    def async_update(self):
        """Update bulb status (if it is available)."""
        _LOGGER.debug("%s async_update", self.who)
        if self.available and self.blocker is None:
            yield from self.refresh_state()

    @asyncio.coroutine
    def stop_effect(self):
        """Stop the currently running effect (if any)."""
        if self.effect_data:
            yield from self.effect_data.effect.async_restore(self)

    @asyncio.coroutine
    def refresh_state(self):
        """Ask the device about its current state and update our copy."""
        msg = yield from AwaitAioLIFX(self).wait(self.device.get_color)
        if msg is not None:
            self.set_power(self.device.power_level)
            self.set_color(*self.device.color)
            self._name = self.device.label

    def find_hsbk(self, **kwargs):
        """Find the desired color from a number of possible inputs."""
        changed_color = False

        hsbk = kwargs.pop(ATTR_HSBK, None)
        if hsbk is not None:
            return [hsbk, True]

        color_name = kwargs.pop(ATTR_COLOR_NAME, None)
        if color_name is not None:
            kwargs[ATTR_RGB_COLOR] = color_util.color_name_to_rgb(color_name)

        if ATTR_RGB_COLOR in kwargs:
            hue, saturation, brightness = \
                convert_rgb_to_hsv(kwargs[ATTR_RGB_COLOR])
            changed_color = True
        else:
            hue = self._hue
            saturation = self._sat
            brightness = self._bri

        if ATTR_BRIGHTNESS in kwargs:
            brightness = kwargs[ATTR_BRIGHTNESS] * (BYTE_MAX + 1)
            changed_color = True
        else:
            brightness = self._bri

        if ATTR_XY_COLOR in kwargs:
            hue, saturation, _ = \
                color_util.color_xy_brightness_to_hsv(
                    *kwargs[ATTR_XY_COLOR],
                    ibrightness=(brightness // (BYTE_MAX + 1)))
            saturation = saturation * (BYTE_MAX + 1)
            changed_color = True

        if ATTR_COLOR_TEMP in kwargs:
            kelvin = int(color_temperature_mired_to_kelvin(
                kwargs[ATTR_COLOR_TEMP]))
            changed_color = True
        else:
            kelvin = self._kel

        return [[hue, saturation, brightness, kelvin], changed_color]

    def set_power(self, power):
        """Set power state value."""
        _LOGGER.debug("set_power: %d", power)
        self._power = (power != 0)

    def set_color(self, hue, sat, bri, kel):
        """Set color state values."""
        self._hue = hue
        self._sat = sat
        self._bri = bri
        self._kel = kel

        red, green, blue = colorsys.hsv_to_rgb(hue / SHORT_MAX,
                                               sat / SHORT_MAX,
                                               bri / SHORT_MAX)

        red = int(red * BYTE_MAX)
        green = int(green * BYTE_MAX)
        blue = int(blue * BYTE_MAX)

        _LOGGER.debug("set_color: %d %d %d %d [%d %d %d]",
                      hue, sat, bri, kel, red, green, blue)

        self._rgb = [red, green, blue]



"""DO NOT MODIFY. Auto-generated by script/fingerprint_frontend."""

FINGERPRINTS = {
    "compatibility.js": "83d9c77748dafa9db49ae77d7f3d8fb0",
    "core.js": "5d08475f03adb5969bd31855d5ca0cfd",
    "frontend.html": "1533f44c55927e814294de757cd7eada",
    "mdi.html": "1cc8593d3684f7f6f3b3854403216f77",
    "micromarkdown-js.html": "93b5ec4016f0bba585521cf4d18dec1a",
    "panels/ha-panel-config.html": "39f00f769faa63ee61f1fe6fc85d67f7",
    "panels/ha-panel-dev-event.html": "2db9c218065ef0f61d8d08db8093cad2",
    "panels/ha-panel-dev-info.html": "61610e015a411cfc84edd2c4d489e71d",
    "panels/ha-panel-dev-service.html": "415552027cb083badeff5f16080410ed",
    "panels/ha-panel-dev-state.html": "d70314913b8923d750932367b1099750",
    "panels/ha-panel-dev-template.html": "567fbf86735e1b891e40c2f4060fec9b",
    "panels/ha-panel-hassio.html": "1d954cfe5f47c4be3cf4f6f5db9a83b2",
    "panels/ha-panel-history.html": "89062c48c76206cad1cec14ddbb1cbb1",
    "panels/ha-panel-iframe.html": "d920f0aa3c903680f2f8795e2255daab",
    "panels/ha-panel-logbook.html": "6dd6a16f52117318b202e60f98400163",
    "panels/ha-panel-map.html": "31c592c239636f91e07c7ac232a5ebc4",
    "panels/ha-panel-zwave.html": "a81f82b48439da80286798558f414a2e",
    "websocket_test.html": "575de64b431fe11c3785bf96d7813450"
}



"""
Component to receive telegram messages.

Either by polling or webhook.
"""

import asyncio
import logging
import voluptuous as vol
import homeassistant.helpers.config_validation as cv
from homeassistant.const import CONF_PLATFORM, CONF_API_KEY
from homeassistant.exceptions import HomeAssistantError
from homeassistant.helpers import discovery, config_per_platform
from homeassistant.setup import async_prepare_setup_platform

DOMAIN = 'telegram_bot'

_LOGGER = logging.getLogger(__name__)

EVENT_TELEGRAM_COMMAND = 'telegram_command'
EVENT_TELEGRAM_TEXT = 'telegram_text'

ATTR_COMMAND = 'command'
ATTR_USER_ID = 'user_id'
ATTR_ARGS = 'args'
ATTR_FROM_FIRST = 'from_first'
ATTR_FROM_LAST = 'from_last'
ATTR_TEXT = 'text'

CONF_ALLOWED_CHAT_IDS = 'allowed_chat_ids'

PLATFORM_SCHEMA = vol.Schema({
    vol.Required(CONF_PLATFORM): cv.string,
    vol.Required(CONF_API_KEY): cv.string,
    vol.Required(CONF_ALLOWED_CHAT_IDS): vol.All(cv.ensure_list,
                                                 [cv.positive_int])
}, extra=vol.ALLOW_EXTRA)


@asyncio.coroutine
def async_setup(hass, config):
    """Setup the telegram bot component."""
    @asyncio.coroutine
    def async_setup_platform(p_type, p_config=None, discovery_info=None):
        """Setup a telegram bot platform."""
        platform = yield from async_prepare_setup_platform(
            hass, config, DOMAIN, p_type)

        if platform is None:
            _LOGGER.error("Unknown notification service specified")
            return

        _LOGGER.info("Setting up %s.%s", DOMAIN, p_type)

        try:
            if hasattr(platform, 'async_setup_platform'):
                notify_service = yield from \
                    platform.async_setup_platform(hass, p_config,
                                                  discovery_info)
            elif hasattr(platform, 'setup_platform'):
                notify_service = yield from hass.loop.run_in_executor(
                    None, platform.setup_platform, hass, p_config,
                    discovery_info)
            else:
                raise HomeAssistantError("Invalid telegram bot platform.")

            if notify_service is None:
                _LOGGER.error(
                    "Failed to initialize telegram bot %s", p_type)
                return

        except Exception:  # pylint: disable=broad-except
            _LOGGER.exception('Error setting up platform %s', p_type)
            return

    setup_tasks = [async_setup_platform(p_type, p_config) for p_type, p_config
                   in config_per_platform(config, DOMAIN)]

    if setup_tasks:
        yield from asyncio.wait(setup_tasks, loop=hass.loop)

    @asyncio.coroutine
    def async_platform_discovered(platform, info):
        """Callback to load a platform."""
        yield from async_setup_platform(platform, discovery_info=info)

    discovery.async_listen_platform(hass, DOMAIN, async_platform_discovered)

    return True


class BaseTelegramBotEntity:
    """The base class for the telegram bot."""

    def __init__(self, hass, allowed_chat_ids):
        """Initialize the bot base class."""
        self.allowed_chat_ids = allowed_chat_ids
        self.hass = hass

    def process_message(self, data):
        """Check for basic message rules and fire an event if message is ok."""
        data = data.get('message')

        if (not data or
                'from' not in data or
                'text' not in data or
                data['from'].get('id') not in self.allowed_chat_ids):
            # Message is not correct.
            _LOGGER.error("Incoming message does not have required data.")
            return False

        event = EVENT_TELEGRAM_COMMAND
        event_data = {
            ATTR_USER_ID: data['from']['id'],
            ATTR_FROM_FIRST: data['from']['first_name'],
            ATTR_FROM_LAST: data['from']['last_name']}

        if data['text'][0] == '/':
            pieces = data['text'].split(' ')
            event_data[ATTR_COMMAND] = pieces[0]
            event_data[ATTR_ARGS] = pieces[1:]

        else:
            event_data[ATTR_TEXT] = data['text']
            event = EVENT_TELEGRAM_TEXT

        self.hass.bus.async_fire(event, event_data)

        return True



"""
Tado component to create a climate device for each zone.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/climate.tado/
"""
import logging

from homeassistant.const import TEMP_CELSIUS
from homeassistant.components.climate import ClimateDevice
from homeassistant.const import ATTR_TEMPERATURE
from homeassistant.components.tado import DATA_TADO

_LOGGER = logging.getLogger(__name__)

CONST_MODE_SMART_SCHEDULE = 'SMART_SCHEDULE'  # Default mytado mode
CONST_MODE_OFF = 'OFF'  # Switch off heating in a zone

# When we change the temperature setting, we need an overlay mode
# wait until tado changes the mode automatic
CONST_OVERLAY_TADO_MODE = 'TADO_MODE'
# the user has change the temperature or mode manually
CONST_OVERLAY_MANUAL = 'MANUAL'
# the temperature will be reset after a timespan
CONST_OVERLAY_TIMER = 'TIMER'

OPERATION_LIST = {
    CONST_OVERLAY_MANUAL: 'Manual',
    CONST_OVERLAY_TIMER: 'Timer',
    CONST_OVERLAY_TADO_MODE: 'Tado mode',
    CONST_MODE_SMART_SCHEDULE: 'Smart schedule',
    CONST_MODE_OFF: 'Off',
}


def setup_platform(hass, config, add_devices, discovery_info=None):
    """Set up the Tado climate platform."""
    tado = hass.data[DATA_TADO]

    try:
        zones = tado.get_zones()
    except RuntimeError:
        _LOGGER.error("Unable to get zone info from mytado")
        return False

    climate_devices = []
    for zone in zones:
        climate_devices.append(create_climate_device(
            tado, hass, zone, zone['name'], zone['id']))

    if len(climate_devices) > 0:
        add_devices(climate_devices, True)
        return True
    else:
        return False


def create_climate_device(tado, hass, zone, name, zone_id):
    """Create a Tado climate device."""
    capabilities = tado.get_capabilities(zone_id)

    unit = TEMP_CELSIUS
    min_temp = float(capabilities['temperatures']['celsius']['min'])
    max_temp = float(capabilities['temperatures']['celsius']['max'])
    ac_mode = capabilities['type'] != 'HEATING'

    data_id = 'zone {} {}'.format(name, zone_id)
    device = TadoClimate(tado,
                         name, zone_id, data_id,
                         hass.config.units.temperature(min_temp, unit),
                         hass.config.units.temperature(max_temp, unit),
                         ac_mode)

    tado.add_sensor(data_id, {
        'id': zone_id,
        'zone': zone,
        'name': name,
        'climate': device
    })

    return device


class TadoClimate(ClimateDevice):
    """Representation of a tado climate device."""

    def __init__(self, store, zone_name, zone_id, data_id,
                 min_temp, max_temp, ac_mode,
                 tolerance=0.3):
        """Initialization of Tado climate device."""
        self._store = store
        self._data_id = data_id

        self.zone_name = zone_name
        self.zone_id = zone_id

        self.ac_mode = ac_mode

        self._active = False
        self._device_is_active = False

        self._unit = TEMP_CELSIUS
        self._cur_temp = None
        self._cur_humidity = None
        self._is_away = False
        self._min_temp = min_temp
        self._max_temp = max_temp
        self._target_temp = None
        self._tolerance = tolerance

        self._current_operation = CONST_MODE_SMART_SCHEDULE
        self._overlay_mode = CONST_MODE_SMART_SCHEDULE

    @property
    def name(self):
        """Return the name of the sensor."""
        return self.zone_name

    @property
    def current_humidity(self):
        """Return the current humidity."""
        return self._cur_humidity

    @property
    def current_temperature(self):
        """Return the sensor temperature."""
        return self._cur_temp

    @property
    def current_operation(self):
        """Return current readable operation mode."""
        return OPERATION_LIST.get(self._current_operation)

    @property
    def operation_list(self):
        """List of available operation modes (readable)."""
        return list(OPERATION_LIST.values())

    @property
    def temperature_unit(self):
        """The unit of measurement used by the platform."""
        return self._unit

    @property
    def is_away_mode_on(self):
        """Return true if away mode is on."""
        return self._is_away

    @property
    def target_temperature(self):
        """Return the temperature we try to reach."""
        return self._target_temp

    def set_temperature(self, **kwargs):
        """Set new target temperature."""
        temperature = kwargs.get(ATTR_TEMPERATURE)
        if temperature is None:
            return

        self._current_operation = CONST_OVERLAY_TADO_MODE
        self._overlay_mode = None
        self._target_temp = temperature
        self._control_heating()

    def set_operation_mode(self, readable_operation_mode):
        """Set new operation mode."""
        operation_mode = CONST_MODE_SMART_SCHEDULE

        for mode, readable in OPERATION_LIST.items():
            if readable == readable_operation_mode:
                operation_mode = mode
                break

        self._current_operation = operation_mode
        self._overlay_mode = None
        self._control_heating()

    @property
    def min_temp(self):
        """Return the minimum temperature."""
        if self._min_temp:
            return self._min_temp
        else:
            # get default temp from super class
            return super().min_temp

    @property
    def max_temp(self):
        """Return the maximum temperature."""
        if self._max_temp:
            return self._max_temp
        else:
            #  Get default temp from super class
            return super().max_temp

    def update(self):
        """Update the state of this climate device."""
        self._store.update()

        data = self._store.get_data(self._data_id)

        if data is None:
            _LOGGER.debug("Recieved no data for zone %s", self.zone_name)
            return

        if 'sensorDataPoints' in data:
            sensor_data = data['sensorDataPoints']
            temperature = float(
                sensor_data['insideTemperature']['celsius'])
            humidity = float(
                sensor_data['humidity']['percentage'])
            setting = 0

            # temperature setting will not exist when device is off
            if 'temperature' in data['setting'] and \
                    data['setting']['temperature'] is not None:
                setting = float(
                    data['setting']['temperature']['celsius'])

            unit = TEMP_CELSIUS

            self._cur_temp = self.hass.config.units.temperature(
                temperature, unit)

            self._target_temp = self.hass.config.units.temperature(
                setting, unit)

            self._cur_humidity = humidity

        if 'tadoMode' in data:
            mode = data['tadoMode']
            self._is_away = mode == 'AWAY'

        if 'setting' in data:
            power = data['setting']['power']
            if power == 'OFF':
                self._current_operation = CONST_MODE_OFF
                self._device_is_active = False
            else:
                self._device_is_active = True

        if 'overlay' in data and data['overlay'] is not None:
            overlay = True
            termination = data['overlay']['termination']['type']
        else:
            overlay = False
            termination = ""

        # If you set mode manualy to off, there will be an overlay
        # and a termination, but we want to see the mode "OFF"

        if overlay and self._device_is_active:
            # There is an overlay the device is on
            self._overlay_mode = termination
            self._current_operation = termination
        else:
            # There is no overlay, the mode will always be
            # "SMART_SCHEDULE"
            self._overlay_mode = CONST_MODE_SMART_SCHEDULE
            self._current_operation = CONST_MODE_SMART_SCHEDULE

    def _control_heating(self):
        """Send new target temperature to mytado."""
        if not self._active and None not in (
                self._cur_temp, self._target_temp):
            self._active = True
            _LOGGER.info("Obtained current and target temperature. "
                         "Tado thermostat active")

        if not self._active or self._current_operation == self._overlay_mode:
            return

        if self._current_operation == CONST_MODE_SMART_SCHEDULE:
            _LOGGER.info("Switching mytado.com to SCHEDULE (default) "
                         "for zone %s", self.zone_name)
            self._store.reset_zone_overlay(self.zone_id)
            self._overlay_mode = self._current_operation
            return

        if self._current_operation == CONST_MODE_OFF:
            _LOGGER.info("Switching mytado.com to OFF for zone %s",
                         self.zone_name)
            self._store.set_zone_overlay(self.zone_id, CONST_OVERLAY_MANUAL)
            self._overlay_mode = self._current_operation
            return

        _LOGGER.info("Switching mytado.com to %s mode for zone %s",
                     self._current_operation, self.zone_name)
        self._store.set_zone_overlay(
            self.zone_id, self._current_operation, self._target_temp)

        self._overlay_mode = self._current_operation



"""
Allows utilizing telegram webhooks.

See https://core.telegram.org/bots/webhooks for details
 about webhooks.

"""
import asyncio
import logging
from ipaddress import ip_network

import voluptuous as vol

from homeassistant.const import (
    EVENT_HOMEASSISTANT_STOP, HTTP_BAD_REQUEST, HTTP_UNAUTHORIZED)
import homeassistant.helpers.config_validation as cv
from homeassistant.components.http import HomeAssistantView
from homeassistant.components.telegram_bot import CONF_ALLOWED_CHAT_IDS, \
    BaseTelegramBotEntity, PLATFORM_SCHEMA
from homeassistant.const import CONF_API_KEY
from homeassistant.components.http.util import get_real_ip

DEPENDENCIES = ['http']
REQUIREMENTS = ['python-telegram-bot==5.3.1']

_LOGGER = logging.getLogger(__name__)

TELEGRAM_HANDLER_URL = '/api/telegram_webhooks'
REMOVE_HANDLER_URL = ''

CONF_TRUSTED_NETWORKS = 'trusted_networks'
DEFAULT_TRUSTED_NETWORKS = [
    ip_network('149.154.167.197/32'),
    ip_network('149.154.167.198/31'),
    ip_network('149.154.167.200/29'),
    ip_network('149.154.167.208/28'),
    ip_network('149.154.167.224/29'),
    ip_network('149.154.167.232/31')
]

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Optional(CONF_TRUSTED_NETWORKS, default=DEFAULT_TRUSTED_NETWORKS):
        vol.All(cv.ensure_list, [ip_network])
})


def setup_platform(hass, config, async_add_devices, discovery_info=None):
    """Setup the polling platform."""
    import telegram
    bot = telegram.Bot(config[CONF_API_KEY])

    current_status = bot.getWebhookInfo()
    handler_url = '{0}{1}'.format(
        hass.config.api.base_url, TELEGRAM_HANDLER_URL)
    if current_status and current_status['url'] != handler_url:
        if bot.setWebhook(handler_url):
            _LOGGER.info("set new telegram webhook %s", handler_url)
        else:
            _LOGGER.error("set telegram webhook failed %s", handler_url)
            return False

    hass.bus.listen_once(
        EVENT_HOMEASSISTANT_STOP,
        lambda event: bot.setWebhook(REMOVE_HANDLER_URL))
    hass.http.register_view(BotPushReceiver(
        hass, config[CONF_ALLOWED_CHAT_IDS], config[CONF_TRUSTED_NETWORKS]))
    return True


class BotPushReceiver(HomeAssistantView, BaseTelegramBotEntity):
    """Handle pushes from telegram."""

    requires_auth = False
    url = TELEGRAM_HANDLER_URL
    name = 'telegram_webhooks'

    def __init__(self, hass, allowed_chat_ids, trusted_networks):
        """Initialize the class."""
        BaseTelegramBotEntity.__init__(self, hass, allowed_chat_ids)
        self.trusted_networks = trusted_networks

    @asyncio.coroutine
    def post(self, request):
        """Accept the POST from telegram."""
        real_ip = get_real_ip(request)
        if not any(real_ip in net for net in self.trusted_networks):
            _LOGGER.warning("Access denied from %s", real_ip)
            return self.json_message('Access denied', HTTP_UNAUTHORIZED)

        try:
            data = yield from request.json()
        except ValueError:
            return self.json_message('Invalid JSON', HTTP_BAD_REQUEST)

        if not self.process_message(data):
            return self.json_message('Invalid message', HTTP_BAD_REQUEST)
        else:
            return self.json({})



"""
Support to interface with the Plex API.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/media_player.plex/
"""
import json
import logging
import os
from datetime import timedelta
from urllib.parse import urlparse

import requests
import voluptuous as vol
from homeassistant import util
from homeassistant.components.media_player import (
    MEDIA_TYPE_MUSIC,
    MEDIA_TYPE_TVSHOW,
    MEDIA_TYPE_VIDEO,
    PLATFORM_SCHEMA,
    SUPPORT_NEXT_TRACK,
    SUPPORT_PAUSE,
    SUPPORT_PLAY,
    SUPPORT_PREVIOUS_TRACK,
    SUPPORT_STOP,
    SUPPORT_TURN_OFF,
    SUPPORT_VOLUME_MUTE,
    SUPPORT_VOLUME_SET,
    MediaPlayerDevice,
)
from homeassistant.const import (
    DEVICE_DEFAULT_NAME,
    STATE_IDLE,
    STATE_OFF,
    STATE_PAUSED,
    STATE_PLAYING,
)
from homeassistant.helpers import config_validation as cv
from homeassistant.helpers.event import track_utc_time_change
from homeassistant.loader import get_component

REQUIREMENTS = ['plexapi==2.0.2']
MIN_TIME_BETWEEN_SCANS = timedelta(seconds=10)
MIN_TIME_BETWEEN_FORCED_SCANS = timedelta(seconds=1)

PLEX_CONFIG_FILE = 'plex.conf'

CONF_INCLUDE_NON_CLIENTS = 'include_non_clients'
CONF_USE_EPISODE_ART = 'use_episode_art'
CONF_USE_CUSTOM_ENTITY_IDS = 'use_custom_entity_ids'
CONF_SHOW_ALL_CONTROLS = 'show_all_controls'

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Optional(CONF_INCLUDE_NON_CLIENTS, default=False):
    cv.boolean,
    vol.Optional(CONF_USE_EPISODE_ART, default=False):
    cv.boolean,
    vol.Optional(CONF_USE_CUSTOM_ENTITY_IDS, default=False):
    cv.boolean,
})

# Map ip to request id for configuring
_CONFIGURING = {}
_LOGGER = logging.getLogger(__name__)


def config_from_file(filename, config=None):
    """Small configuration file management function."""
    if config:
        # We're writing configuration
        try:
            with open(filename, 'w') as fdesc:
                fdesc.write(json.dumps(config))
        except IOError as error:
            _LOGGER.error('Saving config file failed: %s', error)
            return False
        return True
    else:
        # We're reading config
        if os.path.isfile(filename):
            try:
                with open(filename, 'r') as fdesc:
                    return json.loads(fdesc.read())
            except IOError as error:
                _LOGGER.error('Reading config file failed: %s', error)
                # This won't work yet
                return False
        else:
            return {}


def setup_platform(hass, config, add_devices_callback, discovery_info=None):
    """Setup the Plex platform."""
    # get config from plex.conf
    file_config = config_from_file(hass.config.path(PLEX_CONFIG_FILE))

    if len(file_config):
        # Setup a configured PlexServer
        host, token = file_config.popitem()
        token = token['token']
    # Via discovery
    elif discovery_info is not None:
        # Parse discovery data
        host = discovery_info.get('host')
        _LOGGER.info('Discovered PLEX server: %s', host)

        if host in _CONFIGURING:
            return
        token = None
    else:
        return

    setup_plexserver(host, token, hass, config, add_devices_callback)


def setup_plexserver(host, token, hass, config, add_devices_callback):
    """Setup a plexserver based on host parameter."""
    import plexapi.server
    import plexapi.exceptions

    try:
        plexserver = plexapi.server.PlexServer('http://%s' % host, token)
    except (plexapi.exceptions.BadRequest, plexapi.exceptions.Unauthorized,
            plexapi.exceptions.NotFound) as error:
        _LOGGER.info(error)
        # No token or wrong token
        request_configuration(host, hass, config,
                              add_devices_callback)
        return

    # If we came here and configuring this host, mark as done
    if host in _CONFIGURING:
        request_id = _CONFIGURING.pop(host)
        configurator = get_component('configurator')
        configurator.request_done(request_id)
        _LOGGER.info('Discovery configuration done!')

    # Save config
    if not config_from_file(
            hass.config.path(PLEX_CONFIG_FILE), {host: {
                'token': token
            }}):
        _LOGGER.error('failed to save config file')

    _LOGGER.info('Connected to: http://%s', host)

    plex_clients = {}
    plex_sessions = {}
    track_utc_time_change(hass, lambda now: update_devices(), second=30)

    @util.Throttle(MIN_TIME_BETWEEN_SCANS, MIN_TIME_BETWEEN_FORCED_SCANS)
    # pylint: disable=too-many-branches
    def update_devices():
        """Update the devices objects."""
        try:
            devices = plexserver.clients()
        except plexapi.exceptions.BadRequest:
            _LOGGER.exception('Error listing plex devices')
            return
        except OSError:
            _LOGGER.error('Could not connect to plex server at http://%s',
                          host)
            return

        new_plex_clients = []
        for device in devices:
            # For now, let's allow all deviceClass types
            if device.deviceClass in ['badClient']:
                continue

            if device.machineIdentifier not in plex_clients:
                new_client = PlexClient(config, device, None,
                                        plex_sessions, update_devices,
                                        update_sessions)
                plex_clients[device.machineIdentifier] = new_client
                new_plex_clients.append(new_client)
            else:
                plex_clients[device.machineIdentifier].refresh(device, None)

        # add devices with a session and no client (ex. PlexConnect Apple TV's)
        if config.get(CONF_INCLUDE_NON_CLIENTS):
            for machine_identifier, session in plex_sessions.items():
                if (machine_identifier not in plex_clients
                        and machine_identifier is not None):
                    new_client = PlexClient(config, None, session,
                                            plex_sessions, update_devices,
                                            update_sessions)
                    plex_clients[machine_identifier] = new_client
                    new_plex_clients.append(new_client)
                else:
                    plex_clients[machine_identifier].refresh(None, session)

        for machine_identifier, client in plex_clients.items():
            # force devices to idle that do not have a valid session
            if client.session is None:
                client.force_idle()

        if new_plex_clients:
            add_devices_callback(new_plex_clients)

    @util.Throttle(MIN_TIME_BETWEEN_SCANS, MIN_TIME_BETWEEN_FORCED_SCANS)
    def update_sessions():
        """Update the sessions objects."""
        try:
            sessions = plexserver.sessions()
        except plexapi.exceptions.BadRequest:
            _LOGGER.exception('Error listing plex sessions')
            return

        plex_sessions.clear()
        for session in sessions:
            if (session.player is not None and
                    session.player.machineIdentifier is not None):
                plex_sessions[session.player.machineIdentifier] = session

    update_sessions()
    update_devices()


def request_configuration(host, hass, config, add_devices_callback):
    """Request configuration steps from the user."""
    configurator = get_component('configurator')
    # We got an error if this method is called while we are configuring
    if host in _CONFIGURING:
        configurator.notify_errors(_CONFIGURING[host],
                                   'Failed to register, please try again.')

        return

    def plex_configuration_callback(data):
        """The actions to do when our configuration callback is called."""
        setup_plexserver(host,
                         data.get('token'), hass, config,
                         add_devices_callback)

    _CONFIGURING[host] = configurator.request_config(
        hass,
        'Plex Media Server',
        plex_configuration_callback,
        description=('Enter the X-Plex-Token'),
        entity_picture='/static/images/logo_plex_mediaserver.png',
        submit_caption='Confirm',
        fields=[{
            'id': 'token',
            'name': 'X-Plex-Token',
            'type': ''
        }])


# pylint: disable=too-many-instance-attributes, too-many-public-methods
class PlexClient(MediaPlayerDevice):
    """Representation of a Plex device."""

    # pylint: disable=too-many-arguments
    def __init__(self, config, device, session, plex_sessions,
                 update_devices, update_sessions):
        """Initialize the Plex device."""
        from plexapi.utils import NA
        self._app_name = ''
        self._device = None
        self._device_protocol_capabilities = None
        self._is_player_active = False
        self._is_player_available = False
        self._machine_identifier = None
        self._make = ''
        self._media_content_id = None
        self._media_content_rating = None
        self._media_content_type = None
        self._media_duration = None
        self._media_image_url = None
        self._media_title = None
        self._name = None
        self._player_state = 'idle'
        self._previous_volume_level = 1  # Used in fake muting
        self._session = None
        self._session_type = None
        self._session_username = None
        self._state = STATE_IDLE
        self._volume_level = 1  # since we can't retrieve remotely
        self._volume_muted = False  # since we can't retrieve remotely
        self.na_type = NA
        self.config = config
        self.plex_sessions = plex_sessions
        self.update_devices = update_devices
        self.update_sessions = update_sessions

        # Music
        self._media_album_artist = None
        self._media_album_name = None
        self._media_artist = None
        self._media_track = None

        # TV Show
        self._media_episode = None
        self._media_season = None
        self._media_series_title = None

        self.refresh(device, session)

        # Assign custom entity ID if desired
        if self.config.get(CONF_USE_CUSTOM_ENTITY_IDS):
            prefix = ''
            # allow for namespace prefixing when using custom entity names
            if config.get("entity_namespace"):
                prefix = config.get("entity_namespace") + '_'

            # rename the entity id
            if self.machine_identifier:
                self.entity_id = "%s.%s%s" % (
                    'media_player', prefix,
                    self.machine_identifier.lower().replace('-', '_'))
            else:
                if self.name:
                    self.entity_id = "%s.%s%s" % (
                        'media_player', prefix,
                        self.name.lower().replace('-', '_'))

    # pylint: disable=too-many-branches, too-many-statements
    def refresh(self, device, session):
        """Refresh key device data."""
        # new data refresh
        if session:
            self._session = session
        if device:
            self._device = device
            self._session = None

        if self._device:
            self._machine_identifier = self._convert_na_to_none(
                self._device.machineIdentifier)
            self._name = self._convert_na_to_none(
                self._device.title) or DEVICE_DEFAULT_NAME
            self._device_protocol_capabilities = (
                self._device.protocolCapabilities)

        # set valid session, preferring device session
        if self._device and self.plex_sessions.get(
                self._device.machineIdentifier, None):
            self._session = self._convert_na_to_none(self.plex_sessions.get(
                self._device.machineIdentifier, None))

        if self._session:
            self._media_position = self._convert_na_to_none(
                self._session.viewOffset)
            self._media_content_id = self._convert_na_to_none(
                self._session.ratingKey)
            self._media_content_rating = self._convert_na_to_none(
                self._session.contentRating)
        else:
            self._media_position = None
            self._media_content_id = None

        # player dependent data
        if self._session and self._session.player:
            self._is_player_available = True
            self._machine_identifier = self._convert_na_to_none(
                self._session.player.machineIdentifier)
            self._name = self._convert_na_to_none(self._session.player.title)
            self._player_state = self._session.player.state
            self._session_username = self._convert_na_to_none(
                self._session.username)
            self._make = self._convert_na_to_none(self._session.player.device)
        else:
            self._is_player_available = False

        if self._player_state == 'playing':
            self._is_player_active = True
            self._state = STATE_PLAYING
        elif self._player_state == 'paused':
            self._is_player_active = True
            self._state = STATE_PAUSED
        elif self.device:
            self._is_player_active = False
            self._state = STATE_IDLE
        else:
            self._is_player_active = False
            self._state = STATE_OFF

        if self._is_player_active and self._session is not None:
            self._session_type = self._session.type
            self._media_duration = self._convert_na_to_none(
                self._session.duration)
        else:
            self._session_type = None
            self._media_duration = None

        # media type
        if self._session_type == 'clip':
            _LOGGER.debug('Clip content type detected, '
                          'compatibility may vary: %s',
                          self.entity_id)
            self._media_content_type = MEDIA_TYPE_TVSHOW
        elif self._session_type == 'episode':
            self._media_content_type = MEDIA_TYPE_TVSHOW
        elif self._session_type == 'movie':
            self._media_content_type = MEDIA_TYPE_VIDEO
        elif self._session_type == 'track':
            self._media_content_type = MEDIA_TYPE_MUSIC
        else:
            self._media_content_type = None

        # title (movie name, tv episode name, music song name)
        if self._session:
            self._media_title = self._convert_na_to_none(self._session.title)

        # Movies
        if (self.media_content_type == MEDIA_TYPE_VIDEO and
                self._convert_na_to_none(self._session.year) is not None):
            self._media_title += ' (' + str(self._session.year) + ')'

        # TV Show
        if (self._is_player_active and
                self._media_content_type is MEDIA_TYPE_TVSHOW):

            # season number (00)
            if callable(self._convert_na_to_none(self._session.seasons)):
                self._media_season = self._convert_na_to_none(
                    self._session.seasons()[0].index).zfill(2)
            elif self._convert_na_to_none(
                    self._session.parentIndex) is not None:
                self._media_season = self._session.parentIndex.zfill(2)
            else:
                self._media_season = None

            # show name
            self._media_series_title = self._convert_na_to_none(
                self._session.grandparentTitle)

            # episode number (00)
            if self._convert_na_to_none(
                    self._session.index) is not None:
                self._media_episode = str(self._session.index).zfill(2)
        else:
            self._media_season = None
            self._media_series_title = None
            self._media_episode = None

        # Music
        if (self._is_player_active and
                self._media_content_type == MEDIA_TYPE_MUSIC):
            self._media_album_name = self._convert_na_to_none(
                self._session.parentTitle)
            self._media_album_artist = self._convert_na_to_none(
                self._session.grandparentTitle)
            self._media_track = self._convert_na_to_none(self._session.index)
            self._media_artist = self._convert_na_to_none(
                self._session.originalTitle)
            # use album artist if track artist is missing
            if self._media_artist is None:
                _LOGGER.debug(
                    'Using album artist because track artist '
                    'was not found: %s', self.entity_id)
                self._media_artist = self._media_album_artist
        else:
            self._media_album_name = None
            self._media_album_artist = None
            self._media_track = None
            self._media_artist = None

        # set app name to library name
        if (self._session is not None
                and self._session.librarySectionID is not None):
            self._app_name = self._convert_na_to_none(
                self._session.server.library.sectionByID(
                    self._session.librarySectionID).title)
        else:
            self._app_name = ''

        # media image url
        if self._session is not None:
            thumb_url = self._get_thumbnail_url(self._session.thumb)
            if (self.media_content_type is MEDIA_TYPE_TVSHOW
                    and not self.config.get(CONF_USE_EPISODE_ART)):
                thumb_url = self._get_thumbnail_url(
                    self._session.grandparentThumb)

            if thumb_url is None:
                _LOGGER.debug('Using media art because media thumb '
                              'was not found: %s', self.entity_id)
                thumb_url = self._get_thumbnail_url(self._session.art)

            self._media_image_url = thumb_url
        else:
            self._media_image_url = None

    def _get_thumbnail_url(self, property_value):
        """Return full URL (if exists) for a thumbnail property."""
        if self._convert_na_to_none(property_value) is None:
            return None

        if self._session is None or self._session.server is None:
            return None

        url = self._session.server.url(property_value)
        response = requests.get(url, verify=False)
        if response and response.status_code == 200:
            return url

    def force_idle(self):
        """Force client to idle."""
        self._state = STATE_IDLE
        self._session = None

    @property
    def unique_id(self):
        """Return the id of this plex client."""
        return '{}.{}'.format(self.__class__, self.machine_identifier or
                              self.name)

    @property
    def name(self):
        """Return the name of the device."""
        return self._name

    @property
    def machine_identifier(self):
        """Return the machine identifier of the device."""
        return self._machine_identifier

    @property
    def app_name(self):
        """Return the library name of playing media."""
        return self._app_name

    @property
    def device(self):
        """Return the device, if any."""
        return self._device

    @property
    def session(self):
        """Return the session, if any."""
        return self._session

    @property
    def state(self):
        """Return the state of the device."""
        return self._state

    def update(self):
        """Get the latest details."""
        self.update_devices(no_throttle=True)
        self.update_sessions(no_throttle=True)

    # pylint: disable=no-self-use, singleton-comparison
    def _convert_na_to_none(self, value):
        """Convert PlexAPI _NA() instances to None."""
        # PlexAPI will return a "__NA__" object which can be compared to
        # None, but isn't actually None - this converts it to a real None
        # type so that lower layers don't think it's a URL and choke on it
        if value is self.na_type:
            return None
        else:
            return value

    @property
    def _active_media_plexapi_type(self):
        """Get the active media type required by PlexAPI commands."""
        if self.media_content_type is MEDIA_TYPE_MUSIC:
            return 'music'
        else:
            return 'video'

    @property
    def media_content_id(self):
        """Content ID of current playing media."""
        return self._media_content_id

    @property
    def media_content_type(self):
        """Content type of current playing media."""
        if self._session_type == 'clip':
            _LOGGER.debug('Clip content type detected, '
                          'compatibility may vary: %s',
                          self.entity_id)
            return MEDIA_TYPE_TVSHOW
        elif self._session_type == 'episode':
            return MEDIA_TYPE_TVSHOW
        elif self._session_type == 'movie':
            return MEDIA_TYPE_VIDEO
        elif self._session_type == 'track':
            return MEDIA_TYPE_MUSIC
        else:
            return None

    @property
    def media_artist(self):
        """Artist of current playing media, music track only."""
        return self._media_artist

    @property
    def media_album_name(self):
        """Album name of current playing media, music track only."""
        return self._media_album_name

    @property
    def media_album_artist(self):
        """Album artist of current playing media, music track only."""
        return self._media_album_artist

    @property
    def media_track(self):
        """Track number of current playing media, music track only."""
        return self._media_track

    @property
    def media_duration(self):
        """Duration of current playing media in seconds."""
        return self._media_duration

    @property
    def media_image_url(self):
        """Image url of current playing media."""
        return self._media_image_url

    @property
    def media_title(self):
        """Title of current playing media."""
        return self._media_title

    @property
    def media_season(self):
        """Season of curent playing media (TV Show only)."""
        return self._media_season

    @property
    def media_series_title(self):
        """The title of the series of current playing media (TV Show only)."""
        return self._media_series_title

    @property
    def media_episode(self):
        """Episode of current playing media (TV Show only)."""
        return self._media_episode

    @property
    def make(self):
        """The make of the device (ex. SHIELD Android TV)."""
        return self._make

    @property
    def supported_features(self):
        """Flag media player features that are supported."""
        if not self._is_player_active:
            return None

        # force show all controls
        if self.config.get(CONF_SHOW_ALL_CONTROLS):
            return (SUPPORT_PAUSE | SUPPORT_PREVIOUS_TRACK |
                    SUPPORT_NEXT_TRACK | SUPPORT_STOP |
                    SUPPORT_VOLUME_SET | SUPPORT_PLAY |
                    SUPPORT_TURN_OFF | SUPPORT_VOLUME_MUTE)

        # only show controls when we know what device is connecting
        if not self._make:
            return None
        # no mute support
        elif self.make.lower() == "shield android tv":
            _LOGGER.debug(
                'Shield Android TV client detected, disabling mute '
                'controls: %s', self.entity_id)
            return (SUPPORT_PAUSE | SUPPORT_PREVIOUS_TRACK |
                    SUPPORT_NEXT_TRACK | SUPPORT_STOP |
                    SUPPORT_VOLUME_SET | SUPPORT_PLAY |
                    SUPPORT_TURN_OFF)
        # Only supports play,pause,stop (and off which really is stop)
        elif self.make.lower().startswith("tivo"):
            _LOGGER.debug(
                'Tivo client detected, only enabling pause, play, '
                'stop, and off controls: %s', self.entity_id)
            return (SUPPORT_PAUSE | SUPPORT_PLAY | SUPPORT_STOP |
                    SUPPORT_TURN_OFF)
        # Not all devices support playback functionality
        # Playback includes volume, stop/play/pause, etc.
        elif self.device and 'playback' in self._device_protocol_capabilities:
            return (SUPPORT_PAUSE | SUPPORT_PREVIOUS_TRACK |
                    SUPPORT_NEXT_TRACK | SUPPORT_STOP |
                    SUPPORT_VOLUME_SET | SUPPORT_PLAY |
                    SUPPORT_TURN_OFF | SUPPORT_VOLUME_MUTE)
        else:
            return None

    def _local_client_control_fix(self):
        """Detect if local client and adjust url to allow control."""
        if self.device is None:
            return

        # if this device's machineIdentifier matches an active client
        # with a loopback address, the device must be local or casting
        for client in self.device.server.clients():
            if ("127.0.0.1" in client.baseurl and
                    client.machineIdentifier == self.device.machineIdentifier):
                # point controls to server since that's where the
                # playback is occuring
                _LOGGER.debug(
                    'Local client detected, redirecting controls to '
                    'Plex server: %s', self.entity_id)
                server_url = self.device.server.baseurl
                client_url = self.device.baseurl
                self.device.baseurl = "{}://{}:{}".format(
                    urlparse(client_url).scheme,
                    urlparse(server_url).hostname,
                    str(urlparse(client_url).port))

    def set_volume_level(self, volume):
        """Set volume level, range 0..1."""
        if self.device and 'playback' in self._device_protocol_capabilities:
            self._local_client_control_fix()
            self.device.setVolume(
                int(volume * 100), self._active_media_plexapi_type)
            self._volume_level = volume  # store since we can't retrieve

    @property
    def volume_level(self):
        """Return the volume level of the client (0..1)."""
        if (self._is_player_active and self.device and
                'playback' in self._device_protocol_capabilities):
            return self._volume_level

    @property
    def is_volume_muted(self):
        """Return boolean if volume is currently muted."""
        if self._is_player_active and self.device:
            return self._volume_muted

    def mute_volume(self, mute):
        """Mute the volume.

        Since we can't actually mute, we'll:
        - On mute, store volume and set volume to 0
        - On unmute, set volume to previously stored volume
        """
        if not (self.device and
                'playback' in self._device_protocol_capabilities):
            return

        self._volume_muted = mute
        if mute:
            self._previous_volume_level = self._volume_level
            self.set_volume_level(0)
        else:
            self.set_volume_level(self._previous_volume_level)

    def media_play(self):
        """Send play command."""
        if self.device and 'playback' in self._device_protocol_capabilities:
            self._local_client_control_fix()
            self.device.play(self._active_media_plexapi_type)

    def media_pause(self):
        """Send pause command."""
        if self.device and 'playback' in self._device_protocol_capabilities:
            self._local_client_control_fix()
            self.device.pause(self._active_media_plexapi_type)

    def media_stop(self):
        """Send stop command."""
        if self.device and 'playback' in self._device_protocol_capabilities:
            self._local_client_control_fix()
            self.device.stop(self._active_media_plexapi_type)

    def turn_off(self):
        """Turn the client off."""
        # Fake it since we can't turn the client off
        self.media_stop()

    def media_next_track(self):
        """Send next track command."""
        if self.device and 'playback' in self._device_protocol_capabilities:
            self._local_client_control_fix()
            self.device.skipNext(self._active_media_plexapi_type)

    def media_previous_track(self):
        """Send previous track command."""
        if self.device and 'playback' in self._device_protocol_capabilities:
            self._local_client_control_fix()
            self.device.skipPrevious(self._active_media_plexapi_type)

    # pylint: disable=W0613
    def play_media(self, media_type, media_id, **kwargs):
        """Play a piece of media."""
        if not (self.device and
                'playback' in self._device_protocol_capabilities):
            return

        src = json.loads(media_id)

        media = None
        if media_type == 'MUSIC':
            media = self.device.server.library.section(
                src['library_name']).get(src['artist_name']).album(
                    src['album_name']).get(src['track_name'])
        elif media_type == 'EPISODE':
            media = self._get_tv_media(
                src['library_name'], src['show_name'],
                src['season_number'], src['episode_number'])
        elif media_type == 'PLAYLIST':
            media = self.device.server.playlist(src['playlist_name'])
        elif media_type == 'VIDEO':
            media = self.device.server.library.section(
                src['library_name']).get(src['video_name'])

        import plexapi.playlist
        if (media and media_type == 'EPISODE' and
                isinstance(media, plexapi.playlist.Playlist)):
            # delete episode playlist after being loaded into a play queue
            self._client_play_media(media=media, delete=True,
                                    shuffle=src['shuffle'])
        elif media:
            self._client_play_media(media=media, shuffle=src['shuffle'])

    def _get_tv_media(self, library_name, show_name, season_number,
                      episode_number):
        """Find TV media and return a Plex media object."""
        target_season = None
        target_episode = None

        show = self.device.server.library.section(library_name).get(
            show_name)

        if not season_number:
            playlist_name = "{} - {} Episodes".format(
                self.entity_id, show_name)
            return self.device.server.createPlaylist(
                playlist_name, show.episodes())

        for season in show.seasons():
            if int(season.seasonNumber) == int(season_number):
                target_season = season
                break

        if target_season is None:
            _LOGGER.error('Season not found: %s\\%s - S%sE%s', library_name,
                          show_name,
                          str(season_number).zfill(2),
                          str(episode_number).zfill(2))
        else:
            if not episode_number:
                playlist_name = "{} - {} Season {} Episodes".format(
                    self.entity_id, show_name, str(season_number))
                return self.device.server.createPlaylist(
                    playlist_name, target_season.episodes())

            for episode in target_season.episodes():
                if int(episode.index) == int(episode_number):
                    target_episode = episode
                    break

            if target_episode is None:
                _LOGGER.error('Episode not found: %s\\%s - S%sE%s',
                              library_name, show_name,
                              str(season_number).zfill(2),
                              str(episode_number).zfill(2))

        return target_episode

    def _client_play_media(self, media, delete=False, **params):
        """Instruct Plex client to play a piece of media."""
        if not (self.device and
                'playback' in self._device_protocol_capabilities):
            _LOGGER.error('Client cannot play media: %s', self.entity_id)
            return

        import plexapi.playqueue
        playqueue = plexapi.playqueue.PlayQueue.create(self.device.server,
                                                       media, **params)

        # delete dynamic playlists used to build playqueue (ex. play tv season)
        if delete:
            media.delete()

        self._local_client_control_fix()

        server_url = self.device.server.baseurl.split(':')
        self.device.sendCommand('playback/playMedia', **dict({
            'machineIdentifier':
            self.device.server.machineIdentifier,
            'address':
            server_url[1].strip('/'),
            'port':
            server_url[-1],
            'key':
            media.key,
            'containerKey':
            '/playQueues/%s?window=100&own=1' % playqueue.playQueueID,
        }, **params))

    @property
    def device_state_attributes(self):
        """Return the scene state attributes."""
        attr = {}
        attr['media_content_rating'] = self._media_content_rating
        attr['session_username'] = self._session_username
        attr['media_library_name'] = self._app_name

        return attr
